var sk = 'z5YUPHwg1w9bCwknIFkSqSF7bP+08DlPhxCb8vgp';
var ak = '3R5XRVROIOUYS6STJXMQ';
var request = require('request');
var fs = require('fs');
var crypto = require('crypto');
var mime = require('mime');
var path = require('path');
var walk = require('walkdir');
var Retry = require('retryjs');
var logger = require('log4js').getLogger('ksyun');
var async = require('async');
var parser = require('xml2json');

var md5 = function(str) {
  var hash = crypto.createHash('md5');
  hash.update(str);
  return hash.digest('base64');
}

var sign = function(strToSign) {
  var hmac = crypto.createHmac('sha1', sk);
  hmac.update(strToSign);
  var signature = hmac.digest('base64');
  return signature;
};

var sdk = {
  upload: function(options, finish) {
    var buffer;
    var filename;
    var bucket = options.bucket;
    if (options.file) {
      buffer = fs.readFileSync(options.file);
      if (options.filename) {
        filename = options.filename;
      } else if (options.basedir) {
        filename = path.relative(options.basedir, options.file);
      } else {
        filename = path.basename(options.file);
      }
      if (options.dest) {
        filename = options.dest + '/' + filename;
      }
      // console.log(filename);
    }
    var contentMd5 = md5(buffer);
    var contentLength = Buffer.byteLength(buffer);
    var date = new Date().toGMTString();
    var url = 'http://' + bucket + '.kss.ksyun.com/' + filename;
    var pathname = '/' + bucket + '/' + filename;
    var contentType = mime.lookup(filename);
    var strToSign = ['PUT', contentMd5, contentType, date, 'x-kss-acl:public-read', pathname].join('\n');
    var signature = sign(strToSign);

    var retry = new Retry({
      max: 5,
      done: finish,
      fail: finish
    });

    retry.start(function(done, retry) {
      request.put({
        url: url,
        method: 'PUT',
        body: buffer,
        headers: {
          'Content-MD5': contentMd5,
          'Content-Type': contentType,
          'Date': date,
          'x-kss-acl': 'public-read',
          'Authorization': 'KSS ' + ak + ':' + signature
        }
      }, function(err, res, body) {
        if (err) {
          retry();
        } else {
          done(body);
        }
      })
    });
  },
  uploadDir: function(options) {
    var dir = path.resolve(options.dir);
    walk(dir, function(filepath, stat) {
      if (stat.isFile()) {
        sdk.upload({
          file: filepath,
          basedir: dir,
          dest: options.dest,
          service: options.service,
          bucket: options.bucket
        }, function(err) {
          if (err) {
            logger.error(err);
          }
        });
      }
    });
  },
  Multi: function(options) {
    this.filename = options.filename;
    this.bucket = options.bucket;
    this.partNum = 0;
  },
  multiUpload: function(options) {
    var filename;
    var bucket = options.bucket;
    if (options.filename) {
      filename = options.filename;
    } else if (options.basedir) {
      filename = path.relative(options.basedir, options.file);
    } else {
      filename = path.basename(options.file);
    }
    if (options.dest) {
      filename = options.dest + '/' + filename;
    }
    // var url = 'http://' + bucket + '.kss.ksyun.com/' + filename;
    // var contentType = mime.lookup(filename);
    // var date = new Date().toGMTString();
    // var pathname = '/' + bucket + '/' + filename;
    // var uploadId, parts;

    // async.series({
    //   one: initMulti,
    //   two: uploadPart
    // }, function() {
    //   console.log('ok')
    // });

    var multi = new sdk.Multi({
      filename: filename,
      bucket: bucket
    });

    multi.init(function() {
      var offset = 0;
      var size = 10 * 1024 * 1024;
      try {
        var fd = fs.openSync(options.file, 'r');
        var buffer = new Buffer(size);
        var bufflen = fs.readSync(fd, buffer, 0, size, offset);
        buffer = buffer.slice(0, bufflen);
        async.whilst(function() {
          return bufflen !== 0;
        }, function(whileone) {
          multi.uploadPart(buffer, function() {
            offset += bufflen;
            buffer = new Buffer(size);
            bufflen = fs.readSync(fd, buffer, 0, size, offset);
            logger.info(offset / 1024 / 1024 + 'M');
            whileone();
          });
        }, function() {
          multi.listParts(function() {
            multi.end();
          })
        })
      } catch (e) {
        logger.error(e.stack);
      }
    })
  }
};

sdk.Multi.prototype.init = function(callback) {
  var self = this;
  var contentType = mime.lookup(this.filename);
  var date = new Date().toGMTString();
  var pathname = '/' + this.bucket + '/' + this.filename;
  var url = 'http://' + this.bucket + '.kss.ksyun.com/' + this.filename;
  var strToSign = ['POST', contentType, date, 'x-kss-acl:public-read', pathname].join('\n');
  var signature = sign(strToSign);
  var retry = new Retry({
    max: 5,
    done: callback,
    fail: function() {
      logger.info('init error')
    }
  });
  retry.start(function(done, retry) {
    request.post({
      url: url + '?uploads',
      headers: {
        'Content-Type': contentType,
        'Date': date,
        'x-kss-acl': 'public-read',
        'Authorization': 'KSS ' + ak + ':' + signature
      }
    }, function(err, res, body) {
      if (err) {
        retry();
      } else {
        var exp = /<UploadId>(.*?)<\/UploadId>/;
        var result = body.match(exp);
        if (result) {
          self.uploadId = result[1];
          done();
        } else {
          retry();
        }
      }
    })
  });
};

sdk.Multi.prototype.uploadPart = function(buffer, callback) {
  var self = this;
  try {
    var bufflen = buffer.length;
    var contentMd5 = md5(buffer);
    var date = new Date().toGMTString();
    var pathname = '/' + this.bucket + '/' + this.filename;
    var url = 'http://' + this.bucket + '.kss.ksyun.com/' + this.filename;
    var strToSign = ['PUT', date, bufflen, pathname].join('\n');
    var signature = sign(strToSign);
    var retry = new Retry({
      max: 10,
      done: callback,
      fail: function() {
        logger.info('upload part error')
      }
    });
    this.partNum++;
    retry.start(function(done, retry) {
      request.put({
        url: url + '?uploadId=' + self.uploadId + '&partNumber=' + self.partNum,
        body: buffer,
        headers: {
          'Date': date,
          'Content-Length': bufflen,
          // 'Content-MD5': contentMd5,
          'Authorization': 'KSS ' + ak + ':' + signature
        }
      }, function(err, res, body) {
        if (err) {
          logger.error(err);
          retry();
        } else {
          done();
        }
      })
    });

  } catch (e) {
    logger.error(e.stack);
  }
}

sdk.Multi.prototype.listParts = function(callback) {
  var self = this;
  var contentType = mime.lookup(this.filename);
  var date = new Date().toGMTString();
  var pathname = '/' + this.bucket + '/' + this.filename;
  var url = 'http://' + this.bucket + '.kss.ksyun.com/' + this.filename;
  var strToSign = ['GET', date, 'x-kss-acl:public-read', pathname].join('\n');
  var signature = sign(strToSign);
  var retry = new Retry({
    max: 5,
    done: callback,
    fail: function() {
      logger.info('list part error')
    }
  });
  retry.start(function(done, retry) {
    request.get({
      url: url + '?uploadId=' + self.uploadId,
      headers: {
        'Date': date,
        // 'Content-MD5': contentMd5,
        'Authorization': 'KSS ' + ak + ':' + signature
      }
    }, function(err, res, body) {
      if (err) {
        retry();
      } else {
        self.parts = JSON.parse(parser.toJson(body))['ns2:ListPartsResult']['ns2:Part'];
        done();
      }
    })
  });
};

sdk.Multi.prototype.end = function(callback) {
  var self = this;
  var contentType = mime.lookup(this.filename);
  var date = new Date().toGMTString();
  var pathname = '/' + this.bucket + '/' + this.filename;
  var url = 'http://' + this.bucket + '.kss.ksyun.com/' + this.filename;
  var strToSign = ['POST', date, 'x-kss-acl:public-read', pathname].join('\n');
  var signature = sign(strToSign);
  var retry = new Retry({
    max: 5,
    done: callback,
    fail: function() {
      logger.info('end error')
    }
  });
  var requestBody = {
    CompleteMultipartUpload: {
      Part: []
    }
  };
  if (!Array.isArray(this.parts)) {
    requestBody.CompleteMultipartUpload.Part.push({
      PartNumber: {
        $t: this.parts.PartNumber
      },
      ETag: {
        $t: this.parts.ETag
      }
    })
  } else {
    this.parts.forEach(function(item) {
      requestBody.CompleteMultipartUpload.Part.push({
        PartNumber: {
          $t: item.PartNumber
        },
        ETag: {
          $t: item.ETag
        }
      })
    })
  }
  retry.start(function(done, retry) {
    request.post({
      url: url + '?uploadId=' + self.uploadId,
      body: parser.toXml(requestBody, {
        arrayNotation: true
      }),
      headers: {
        'Date': date,
        'x-kss-acl': 'public-read',
        'Authorization': 'KSS ' + ak + ':' + signature
      }
    }, function(err, res, body) {
      if (err) {
        retry();
      } else {
        done()
      }
    })
  });
}

// sdk.upload({
//   file: 'package.json'
// }, function(err) {
//   console.error(err);
// });

// sdk.multiUpload({
//   file: 'xj1.mp4',
//   // file: 'pzxyygp.mp4',
//   bucket: 'movie'
// })

// sdk.multiUpload({
//   file: 'xj2.mp4',
//   // file: 'pzxyygp.mp4',
//   bucket: 'movie'
// })

module.exports = sdk;